/***********************************************************************************************************************
 *
 * blueBill Mobile - open source birdwatching
 * ==========================================
 *
 * Copyright (C) 2009, 2010 by Tidalwave s.a.s. (http://www.tidalwave.it)
 * http://bluebill.tidalwave.it/mobile/
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: Taxonomy.java,v 52aa5de73d43 2010/08/17 22:21:53 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.taxonomy;

import javax.annotation.Nonnull;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.HashSet;
import java.util.Set;
import java.io.Serializable;
import it.tidalwave.util.Id;
import it.tidalwave.util.NotFoundException;
import java.util.Arrays;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: $
 *
 **********************************************************************************************************************/
public class Taxonomy implements Serializable
  {
    private static final long serialVersionUID = 65968939768498676L;

    protected final Map<Locale, LocalizedNamesBag> namesBagByLocale = new HashMap<Locale, LocalizedNamesBag>();

    protected final Map<String, List<String>> alternateLanguageCodesMap = new HashMap<String, List<String>>();

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public static enum Type implements Serializable
      {
        CLASS, ORDER, FAMILY, GENUS, SPECIES, SUBSPECIES;

        private static final long serialVersionUID = 6079865922121355L;

        @Nonnull
        public Type getNarrower()
          throws NotFoundException
          {
            final int narrowerIndex = this.ordinal() + 1;

            if (narrowerIndex >= values().length)
              {
                throw new NotFoundException("No narrower for " + this);
              }

            return values()[narrowerIndex];
          }
      }

    private final String displayName;

    protected final Map<Type, List<Taxon>> taxonListMapByType = new HashMap<Type, List<Taxon>>();

    protected final Map<Id, Taxon> taxonMapById = new HashMap<Id, Taxon>();

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public Taxonomy (final @Nonnull String displayName)
      {
        this.displayName = displayName;
        
        alternateLanguageCodesMap.put("en",    Arrays.asList("en_gb", "en_us", "en_ca"));
        alternateLanguageCodesMap.put("en_gb", Arrays.asList("en",    "en_us", "en_ca"));
        alternateLanguageCodesMap.put("en_us", Arrays.asList("en_ca", "en",    "en_gb"));
        alternateLanguageCodesMap.put("en_ca", Arrays.asList("en_us", "en",    "en_gb"));

        alternateLanguageCodesMap.put("fr",    Arrays.asList("fr_fr", "fr_be", "fr_ca"));
        alternateLanguageCodesMap.put("fr_fr", Arrays.asList("fr",    "fr_be", "fr_ca"));
        alternateLanguageCodesMap.put("fr_be", Arrays.asList("fr_fr", "fr",    "fr_ca"));
        alternateLanguageCodesMap.put("fr_ca", Arrays.asList("fr",    "fr_fr", "fr_be"));

        alternateLanguageCodesMap.put("it",    Arrays.asList("it_it", "it_ch"));
        alternateLanguageCodesMap.put("it_it", Arrays.asList("it",    "it_ch"));
        alternateLanguageCodesMap.put("it_ch", Arrays.asList("it",    "it_it"));
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public String getDisplayName()
      {
        return displayName;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Taxon findTaxonById (final @Nonnull Id id)
      throws NotFoundException
      {
        return NotFoundException.throwWhenNull(taxonMapById.get(id), "Can't find Taxon with id=" + id);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public List<Taxon> findTaxaByType (final @Nonnull Type type)
      {
        return Collections.unmodifiableList(taxonListMapByType.get(type));
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    protected String getDisplayName (final @Nonnull Taxon taxon, @Nonnull Locale locale)
      throws NotFoundException
      {
        final String languageCode = getLanguageCode(locale);

//        System.err.println("  " + locale + " " + languageCode);
        // namesBagByLocale.containsKey(locale) seems not to work

        try
          {
            return tryDisplayName(taxon, languageCode);
          }
        catch (NotFoundException e)
          {
            final List<String> altLanguageCodes = alternateLanguageCodesMap.get(languageCode);

            if (altLanguageCodes != null)
              {
                for (final String altLanguageCode : altLanguageCodes)
                  {
                    try
                      {
                        return tryDisplayName(taxon, altLanguageCode);
                      }
                    catch (NotFoundException e2)
                      {
                        // next round
                      }
                  }
              }
 
            else if (languageCode.length() == 5) // e.g. es_ES -> es
              {
                return tryDisplayName(taxon, languageCode.substring(0, 2));
              }

            throw e;
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private String tryDisplayName (final @Nonnull Taxon taxon, final @Nonnull String languageCode)
      throws NotFoundException
      {
        final Locale locale = TaxonomyFactory.LOCALE_MAP_BY_CODE.get(languageCode);

        if (locale == null)
          {
            throw new NotFoundException();
          }

        return NotFoundException.throwWhenNull(namesBagByLocale.get(locale), "").getLocalizedName(taxon);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Set<Locale> getLocales()
      {
        return new HashSet<Locale>(namesBagByLocale.keySet());
      }

//    /*******************************************************************************************************************
//     *
//     *
//     ******************************************************************************************************************/
//    @Nonnull
//    protected static String findAlternateLanguage (final @Nonnull String language)
//      {
//        String newLanguage = language;
//
//        if (newLanguage.length() == 5)
//          {
//            if (newLanguage.equalsIgnoreCase("en_ca"))
//              {
//                newLanguage = "en_us";
//              }
//            else
//              {
//                newLanguage = newLanguage.substring(0, 2);
//              }
//          }
//        else if (newLanguage.equals("en"))
//          {
//            newLanguage += "_us";
//          }
//        else if (newLanguage.equals("fr"))
//          {
//            newLanguage += "_ca";
//          }
//        //            else
//            // Don't fail now; sometimes the language might be fine.
//        //              {
//        //                  System.err.println("Cannot normalize locale: " + newLanguage);
//        //                throw new NotFoundException();
//        //              }
//
////        System.err.println("LOCALE CHANGED FROM " + language + " TO " + newLanguage);
//        return newLanguage;
//      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private static String getLanguageCode (final @Nonnull Locale locale)
      {
        String languageCode = locale.getLanguage().toLowerCase();
        final String countryCode = locale.getCountry().toLowerCase();

        if ((countryCode != null) && !countryCode.equals(""))
          {
            languageCode += "_" + countryCode;
          }
        
        return languageCode;
      }
  }
